@@ -120,7 +120,7 @@ void *client(void *arg) {
     cleanup(c, do_client(c));
 #ifdef USE_FORK
     if(!c->opt->option.remote) /* 'exec' specified */
-        exec_status(); /* null SIGCHLD handler was used */
+        child_status(); /* null SIGCHLD handler was used */
 #else
     enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
     s_log(LOG_DEBUG, "%s finished (%d left)", c->opt->servname,
@@ -172,7 +172,7 @@ static int do_client(CLI *c) {
 
 static int init_local(CLI *c) {
     SOCKADDR_UNION addr;
-    int addrlen;
+    socklen_t addrlen;
 
     addrlen=sizeof(SOCKADDR_UNION);
     if(getpeername(c->local_rfd.fd, &addr.sa, &addrlen)<0) {
@@ -387,14 +387,14 @@ static int transfer(CLI *c) {
         s_poll_zero(&c->fds); /* Initialize the structure */
         if(sock_rd && c->sock_ptr<BUFFSIZE) /* socket input buffer not full*/
             s_poll_add(&c->fds, c->sock_rfd->fd, 1, 0);
-        if((ssl_rd && c->ssl_ptr<BUFFSIZE) /* SSL input buffer not full */ ||
+        if((ssl_rd && c->ssl_ptr<BUFFSIZE) || /* SSL input buffer not full */
                 ((c->sock_ptr || ssl_closing==CL_RETRY) && want_rd))
                 /* want to SSL_write or SSL_shutdown but read from the
                  * underlying socket needed for the SSL protocol */
             s_poll_add(&c->fds, c->ssl_rfd->fd, 1, 0);
         if(c->ssl_ptr) /* SSL input buffer not empty */
             s_poll_add(&c->fds, c->sock_wfd->fd, 0, 1);
-        if(c->sock_ptr /* socket input buffer not empty */ ||
+        if(c->sock_ptr || /* socket input buffer not empty */
                 ssl_closing==CL_INIT /* need to send close_notify */ ||
                 ((c->ssl_ptr<BUFFSIZE || ssl_closing==CL_RETRY) && want_wr))
                 /* want to SSL_read or SSL_shutdown but write to the
@@ -465,10 +465,8 @@ static int transfer(CLI *c) {
         }
 
         /****************************** write to SSL */
-        if(ssl_wr && ( /* SSL is open for write */
-                (c->sock_ptr && ssl_can_wr) ||
-                /* application data can be written */
-                (want_rd && ssl_can_rd)
+        if(ssl_wr && c->sock_ptr && ( /* output buffer not empty */
+                ssl_can_wr || (want_rd && ssl_can_rd)
                 /* SSL_write wants to read from the underlying descriptor */
                 )) {
             num=SSL_write(c->ssl, c->sock_buff, c->sock_ptr);
@@ -526,10 +524,8 @@ static int transfer(CLI *c) {
         }
 
         /****************************** read from SSL */
-        if(ssl_rd && ( /* SSL is open for read */
-                (c->ssl_ptr<BUFFSIZE && ssl_can_rd) ||
-                /* there's any application data coming in */
-                (want_wr && ssl_can_wr) ||
+        if(ssl_rd && c->ssl_ptr<BUFFSIZE  && ( /* input buffer not full */
+                ssl_can_rd || (want_wr && ssl_can_wr) ||
                 /* SSL_read wants to write to the underlying descriptor */
                 (check_SSL_pending && SSL_pending(c->ssl))
                 /* write made space from full buffer */
@@ -622,8 +618,8 @@ static int transfer(CLI *c) {
                 ssl_can_rd ? "yes" : "no", ssl_can_wr ? "yes" : "no");
             s_log(LOG_ERR, "ssl want: rd=%s wr=%s",
                 want_rd ? "yes" : "no", want_wr ? "yes" : "no");
-            s_log(LOG_ERR, "socket write buffer: %d byte(s), "
-                "ssl write buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
+            s_log(LOG_ERR, "socket input buffer: %d byte(s), "
+                "ssl input buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
             s_log(LOG_ERR, "check_SSL_pending=%d, ssl_closing=%d",
                 check_SSL_pending, ssl_closing);
             return -1;
@@ -641,6 +637,7 @@ static int parse_socket_error(const char *text) {
         return 0;
     case EWOULDBLOCK:
         s_log(LOG_NOTICE, "%s would block: retrying", text);
+        sleep(1); /* Microsoft bug KB177346 */
         return 0;
 #if EAGAIN!=EWOULDBLOCK
     case EAGAIN:
@@ -702,21 +699,49 @@ static void print_cipher(CLI *c) { /* print negotiated cipher */
 static int auth_libwrap(CLI *c) {
 #ifdef USE_LIBWRAP
     struct request_info request;
-    int result;
+    int fd[2];
+    int result=0; /* deny by default */
 
-    enter_critical_section(CRIT_INET); /* libwrap is not mt-safe */
-    request_init(&request,
-        RQ_DAEMON, c->opt->servname, RQ_FILE, c->local_rfd.fd, 0);
-    fromhost(&request);
-    result=hosts_access(&request);
-    leave_critical_section(CRIT_INET);
+    if(pipe(fd)<0) {
+        ioerror("pipe");
+        return -1;
+    }
+    if(alloc_fd(fd[0]) || alloc_fd(fd[1]))
+        return -1;
+    switch(fork()) {
+    case -1:    /* error */
+        close(fd[0]);
+        close(fd[1]);
+        ioerror("fork");
+        return -1;
+    case  0:    /* child */
+        close(fd[0]); /* read side */
+        request_init(&request,
+            RQ_DAEMON, c->opt->servname, RQ_FILE, c->local_rfd.fd, 0);
+        fromhost(&request);
+        result=hosts_access(&request);
+        write_blocking(c, fd[1], (u8 *)&result, sizeof(result));
+            /* ignore the returned error */
+        close(fd[1]); /* write side */
+        _exit(0);
+    default:    /* parent */
+        close(fd[1]); /* write side */
+        read_blocking(c, fd[0], (u8 *)&result, sizeof(result));
+            /* ignore the returned error */
+        close(fd[0]); /* read side */
+        /* no need to wait() for zombies here:
+         *  - in UCONTEXT/PTHREAD mode they're removed using the signal pipe
+         *  - in FORK mode they're removed with the client process */
+    }
 
-    if (!result) {
+    if(!result) {
         s_log(LOG_WARNING, "Connection from %s REFUSED by libwrap",
             c->accepting_address);
-        s_log(LOG_DEBUG, "See hosts_access(5) for details");
+        s_log(LOG_DEBUG, "See hosts_access(5) manual for details");
         return -1; /* FAILED */
     }
+    s_log(LOG_DEBUG, "Connection from %s permitted by libwrap",
+        c->accepting_address);
 #endif
     return 0; /* OK */
 }
@@ -867,7 +892,7 @@ static int connect_local(CLI *c) { /* spawn local process */
 static int make_sockets(int fd[2]) { /* make a pair of connected sockets */
 #ifdef INET_SOCKET_PAIR
     SOCKADDR_UNION addr;
-    int addrlen;
+    socklen_t addrlen;
     int s; /* temporary socket awaiting for connection */
 
     if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
@@ -976,7 +1001,8 @@ static int connect_remote(CLI *c) { /* connect to remote host */
 
     /* wait for the result of a non-blocking connect */
 static int connect_wait(CLI *c, int fd, int timeout) {
-    int error, len;
+    int error;
+    socklen_t optlen;
 
     s_log(LOG_DEBUG, "connect_wait: waiting %d seconds", timeout);
     s_poll_zero(&c->fds);
@@ -992,8 +1018,8 @@ static int connect_wait(CLI *c, int fd, int timeout) {
         if(s_poll_canread(&c->fds, fd)) {
             /* just connected socket should not be ready for read */
             /* get the resulting error code, now */
-            len=sizeof(error);
-            if(!getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&error, &len))
+            optlen=sizeof(error);
+            if(!getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&error, &optlen))
                 errno=error;
             if(errno) { /* really an error? */
                 sockerror("connect_wait: getsockopt");
