@@ -31,7 +31,7 @@
 #include "common.h"
 #include "prototypes.h"
 
-/* #define DEBUG_UCONTEXT */
+#define DEBUG_UCONTEXT
 
 #ifndef USE_WIN32
 static int signal_pipe[2]={-1, -1};
@@ -45,7 +45,7 @@ static void client_status(void); /* dead children detected */
 
 static void setnonblock(int, unsigned long);
 
-#ifdef HAVE_POLL
+#ifdef USE_POLL
 
 void s_poll_zero(s_poll_set *fds) {
     fds->nfds=0;
@@ -77,7 +77,7 @@ int s_poll_canread(s_poll_set *fds, int fd) {
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLIN;
+            return fds->ufds[i].revents&~POLLOUT; /* read or error */
     return 0;
 }
 
@@ -86,7 +86,7 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLOUT;
+            return fds->ufds[i].revents&POLLOUT; /* write */
     return 0;
 }
 
@@ -156,14 +156,17 @@ static void scan_waiting_queue(void) {
         for(i=0; i<ctx->fds->nfds; i++) {
             ctx->fds->ufds[i].revents=ufds[nfds].revents;
 #ifdef DEBUG_UCONTEXT
-            s_log(LOG_DEBUG, "CONTEXT %ld, FD=%d, (%s%s)->(%s%s)",
+            s_log(LOG_DEBUG, "CONTEXT %ld, FD=%d, (%s%s)->(%s%s%s%s%s)",
                 ctx->id, ufds[nfds].fd,
                 ufds[nfds].events & POLLIN ? "IN" : "",
                 ufds[nfds].events & POLLOUT ? "OUT" : "",
                 ufds[nfds].revents & POLLIN ? "IN" : "",
-                ufds[nfds].revents & POLLOUT ? "OUT" : "");
+                ufds[nfds].revents & POLLOUT ? "OUT" : "",
+                ufds[nfds].revents & POLLERR ? "ERR" : "",
+                ufds[nfds].revents & POLLHUP ? "HUP" : "",
+                ufds[nfds].revents & POLLNVAL ? "NVAL" : "");
 #endif
-            if(ufds[nfds].revents & (POLLIN|POLLOUT))
+            if(ufds[nfds].revents)
                 ctx->ready++;
             nfds++;
         }
@@ -310,14 +313,27 @@ int s_poll_wait(s_poll_set *fds, int timeout) {
     return retval;
 }
 
-#endif /* HAVE_POLL */
+#endif /* USE_POLL */
 
 #ifndef USE_WIN32
 
 static void sigchld_handler(int sig) { /* SIGCHLD detected */
-    int save_errno=errno;
+    int save_errno;
+#ifdef __sgi
+    int status;
+#endif
 
+    save_errno=errno;
+#ifdef __sgi
+    while(wait_for_pid(-1, &status, WNOHANG)>0) {
+        /* no logging is possible in a signal handler */
+#ifdef USE_FORK
+        num_clients--; /* one client less */
+#endif /* USE_FORK */
+    }
+#else /* __sgi */
     write(signal_pipe[1], signal_buffer, 1);
+#endif /* __sgi */
     signal(SIGCHLD, sigchld_handler);
     errno=save_errno;
 }
@@ -339,38 +355,15 @@ int signal_pipe_init(void) {
 }
 
 static void signal_pipe_empty(void) {
+    s_log(LOG_DEBUG, "Cleaning up the signal pipe");
     read(signal_pipe[0], signal_buffer, sizeof(signal_buffer));
-#ifdef USE_PTHREAD
-    exec_status(); /* report status of 'exec' process */
-#endif /* USE_PTHREAD */
 #ifdef USE_FORK
     client_status(); /* report status of client process */
+#else /* USE_UCONTEXT || USE_PTHREAD */
+    child_status();  /* report status of libwrap or 'exec' process */
 #endif /* defined USE_FORK */
 }
 
-void exec_status(void) { /* dead local ('exec') process detected */
-    int pid, status;
-
-#ifdef HAVE_WAIT_FOR_PID
-    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
-#else
-    if((pid=wait(&status))>0) {
-#endif
-#ifdef WIFSIGNALED
-        if(WIFSIGNALED(status)) {
-            s_log(LOG_INFO, "Local process %d terminated on signal %d",
-                pid, WTERMSIG(status));
-        } else {
-            s_log(LOG_INFO, "Local process %d finished with code %d",
-                pid, WEXITSTATUS(status));
-        }
-#else
-        s_log(LOG_INFO, "Local process %d finished with status %d",
-            pid, status);
-#endif
-    }
-}
-
 #ifdef USE_FORK
 static void client_status(void) { /* dead children detected */
     int pid, status;
@@ -399,6 +392,29 @@ static void client_status(void) { /* dead children detected */
 }
 #endif /* defined USE_FORK */
 
+void child_status(void) { /* dead libwrap or 'exec' process detected */
+    int pid, status;
+
+#ifdef HAVE_WAIT_FOR_PID
+    while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
+#else
+    if((pid=wait(&status))>0) {
+#endif
+#ifdef WIFSIGNALED
+        if(WIFSIGNALED(status)) {
+            s_log(LOG_INFO, "Child process %d terminated on signal %d",
+                pid, WTERMSIG(status));
+        } else {
+            s_log(LOG_INFO, "Child process %d finished with code %d",
+                pid, WEXITSTATUS(status));
+        }
+#else
+        s_log(LOG_INFO, "Child process %d finished with status %d",
+            pid, status);
+#endif
+    }
+}
+
 #endif /* !defined USE_WIN32 */
 
 int alloc_fd(int sock) {
@@ -425,7 +441,7 @@ static void setnonblock(int sock, unsigned long l) {
     do {
         flags=fcntl(sock, F_GETFL, 0);
     }while(flags<0 && get_last_socket_error()==EINTR);
-    flags=l ? flags|O_NONBLOCK : flags&(~O_NONBLOCK);
+    flags=l ? flags|O_NONBLOCK : flags&~O_NONBLOCK;
     do {
         retval=fcntl(sock, F_SETFL, flags);
     }while(retval<0 && get_last_socket_error()==EINTR);
