@@ -31,6 +31,8 @@
 #include "common.h"
 #include "prototypes.h"
 
+#define CONFLINELEN (16*1024)
+
 static int parse_debug_level(char *);
 static int parse_ssl_option(char *);
 static int print_socket_options(void);
@@ -52,7 +54,8 @@ typedef enum {
     CMD_HELP
 } CMD;
 
-static char *option_not_found="Specified option name is not valid here";
+static char *option_not_found=
+    "Specified option name is not valid here";
 
 static char *global_options(CMD cmd, char *opt, char *arg) {
 
@@ -989,21 +992,23 @@ static void syntax(char *confname) {
     log_raw(" ");
     log_raw("Syntax:");
 #ifdef USE_WIN32
-    log_raw("stunnel [filename] | -help | -version | -sockets"
-            " | -install | -uninstall");
+    log_raw("stunnel [ [-install | -uninstall] [-quiet] [<filename>] ]"
+        " | -help | -version | -sockets");
 #else
-    log_raw("stunnel [filename] | -fd [n] | -help | -version | -sockets");
+    log_raw("stunnel [<filename>] | -fd <n> | -help | -version | -sockets");
 #endif
-    log_raw("    filename    - use specified config file instead of %s",
+    log_raw("    <filename>  - use specified config file instead of %s",
         confname);
-    log_raw("    -fd n       - read the config file from specified file descriptor");
-    log_raw("    -help       - get config file help");
-    log_raw("    -version    - display version and defaults");
-    log_raw("    -sockets    - display default socket options");
 #ifdef USE_WIN32
     log_raw("    -install    - install NT service");
     log_raw("    -uninstall  - uninstall NT service");
+    log_raw("    -quiet      - don't display a message box on success");
+#else
+    log_raw("    -fd <n>     - read the config file from a file descriptor");
 #endif
+    log_raw("    -help       - get config file help");
+    log_raw("    -version    - display version and defaults");
+    log_raw("    -sockets    - display default socket options");
     exit(1);
 }
 
@@ -1014,7 +1019,7 @@ void parse_config(char *name, char *parameter) {
     char *default_config_file="stunnel.conf";
 #endif
     FILE *fp;
-    char line[STRLEN], *arg, *opt, *errstr, *filename;
+    char confline[CONFLINELEN], *arg, *opt, *errstr, *filename;
     int line_number, i;
     LOCAL_OPTIONS *section, *new_section;
     
@@ -1069,7 +1074,7 @@ void parse_config(char *name, char *parameter) {
         fp=fopen(name, "r");
         if(!fp) {
 #ifdef USE_WIN32
-            /* Win32 doesn't seem to set errno in fopen() */
+            /* fopen() does not return the error via GetLastError() on Win32 */
             log_raw("Failed to open configuration file %s", name);
 #else
             ioerror(name);
@@ -1079,14 +1084,14 @@ void parse_config(char *name, char *parameter) {
         filename=name;
     }
     line_number=0;
-    while(fgets(line, STRLEN, fp)) {
+    while(fgets(confline, CONFLINELEN, fp)) {
         line_number++;
-        opt=line;
-        while(isspace(*opt))
+        opt=confline;
+        while(isspace((unsigned char)*opt))
             opt++; /* remove initial whitespaces */
-        for(i=strlen(opt)-1; i>=0 && isspace(opt[i]); i--)
+        for(i=strlen(opt)-1; i>=0 && isspace((unsigned char)opt[i]); i--)
             opt[i]='\0'; /* remove trailing whitespaces */
-        if(opt[0]=='\0' || opt[0]=='#' || opt[0]==';') /* empty line or comment */
+        if(opt[0]=='\0' || opt[0]=='#' || opt[0]==';') /* empty or comment */
             continue;
         if(opt[0]=='[' && opt[strlen(opt)-1]==']') { /* new section */
             errstr=section_validate(section);
@@ -1109,15 +1114,15 @@ void parse_config(char *name, char *parameter) {
             section=new_section;
             continue;
         }
-        arg=strchr(line, '=');
+        arg=strchr(confline, '=');
         if(!arg) {
             log_raw("file %s line %d: No '=' found", filename, line_number);
             exit(1);
         }
         *arg++='\0'; /* split into option name and argument value */
-        for(i=strlen(opt)-1; i>=0 && isspace(opt[i]); i--)
+        for(i=strlen(opt)-1; i>=0 && isspace((unsigned char)opt[i]); i--)
             opt[i]='\0'; /* remove trailing whitespaces */
-        while(isspace(*arg))
+        while(isspace((unsigned char)*arg))
             arg++; /* remove initial whitespaces */
         errstr=service_options(CMD_EXEC, section, opt, arg);
         if(section==&local_options && errstr==option_not_found)
@@ -1133,10 +1138,16 @@ void parse_config(char *name, char *parameter) {
         exit(1);
     }
     fclose(fp);
-    if(!local_options.next && section->option.accept) {
-        log_raw("accept option is not allowed in inetd mode");
-        log_raw("remove accept option or define a [section]");
-        exit(1);
+    if(!local_options.next) { /* inetd mode */
+        if (section->option.accept) {
+            log_raw("accept option is not allowed in inetd mode");
+            log_raw("remove accept option or define a [section]");
+            exit(1);
+        }
+        if (!section->option.remote && !section->execname) {
+            log_raw("inetd mode must define a remote host or an executable");
+            exit(1);
+        }
     }
     if(!options.option.client)
         options.option.cert=1; /* Server always needs a certificate */
@@ -1185,9 +1196,9 @@ static char **argalloc(char *str) { /* Allocate 'exec' argumets */
     i=0;
     while(*ptr && i<max_arg) {
         retval[i++]=ptr;
-        while(*ptr && !isspace(*ptr))
+        while(*ptr && !isspace((unsigned char)*ptr))
             ptr++;
-        while(*ptr && isspace(*ptr))
+        while(*ptr && isspace((unsigned char)*ptr))
             *ptr++='\0';
     }
     retval[i]=NULL;
@@ -1368,7 +1379,8 @@ SOCK_OPT sock_opts[] = {
 };
 
 static int print_socket_options(void) {
-    int fd, len;
+    int fd;
+    socklen_t optlen;
     SOCK_OPT *ptr;
     OPT_UNION val;
     char line[STRLEN];
@@ -1386,8 +1398,9 @@ static int print_socket_options(void) {
         print_option(line, ptr->opt_type, ptr->opt_val[1]);
         print_option(line, ptr->opt_type, ptr->opt_val[2]);
         /* display OS default value */
-        len = sizeof(val);
-        if(getsockopt(fd, ptr->opt_level, ptr->opt_name, (void *)&val, &len)) {
+        optlen=sizeof(val);
+        if(getsockopt(fd, ptr->opt_level,
+                ptr->opt_name, (void *)&val, &optlen)) {
             if(get_last_socket_error()!=ENOPROTOOPT) {
                 log_raw("%s", line); /* dump the name and assigned values */
                 sockerror("getsockopt");
